@@ -2,7 +2,7 @@
 
 Copyright (c) 1995, 2016, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2008, Google Inc.
-Copyright (c) 2017, 2019, MariaDB Corporation.
+Copyright (c) 2017, 2020, MariaDB Corporation.
 
 Portions of this file contain modifications contributed and copyrighted by
 Google, Inc. Those modifications are gratefully acknowledged and are described
@@ -202,9 +202,6 @@ rw_lock_create_func(
 	new(lock) rw_lock_t();
 #endif /* UNIV_DEBUG */
 
-	/* If this is the very first time a synchronization object is
-	created, then the following call initializes the sync system. */
-
 	lock->lock_word = X_LOCK_DECR;
 	lock->waiters = 0;
 
@@ -238,9 +235,7 @@ rw_lock_create_func(
 	lock->is_block_lock = 0;
 
 	mutex_enter(&rw_lock_list_mutex);
-
 	UT_LIST_ADD_FIRST(rw_lock_list, lock);
-
 	mutex_exit(&rw_lock_list_mutex);
 }
 
@@ -254,8 +249,7 @@ rw_lock_free_func(
 	rw_lock_t*	lock)	/*!< in/out: rw-lock */
 {
 	ut_ad(rw_lock_validate(lock));
-	ut_a(my_atomic_load32_explicit(&lock->lock_word,
-				       MY_MEMORY_ORDER_RELAXED) == X_LOCK_DECR);
+	ut_a(lock->lock_word == X_LOCK_DECR);
 
 	mutex_enter(&rw_lock_list_mutex);
 
@@ -292,13 +286,15 @@ rw_lock_s_lock_spin(
 
 	ut_ad(rw_lock_validate(lock));
 
+	rw_lock_stats.rw_s_spin_wait_count.inc();
+
 lock_loop:
 
 	/* Spin waiting for the writer field to become free */
 	HMT_low();
+	ulint j = i;
 	while (i < srv_n_spin_wait_rounds &&
-	       my_atomic_load32_explicit(&lock->lock_word,
-					 MY_MEMORY_ORDER_RELAXED) <= 0) {
+	       lock->lock_word <= 0) {
 		ut_delay(srv_spin_wait_delay);
 		i++;
 	}
@@ -308,7 +304,7 @@ lock_loop:
 		os_thread_yield();
 	}
 
-	++spin_count;
+	spin_count += lint(i - j);
 
 	/* We try once again to obtain the lock */
 	if (rw_lock_s_lock_low(lock, pass, file_name, line)) {
@@ -338,7 +334,7 @@ lock_loop:
 
 		/* Set waiters before checking lock_word to ensure wake-up
 		signal is sent. This may lead to some unnecessary signals. */
-		my_atomic_fas32_explicit(&lock->waiters, 1, MY_MEMORY_ORDER_ACQUIRE);
+		lock->waiters.exchange(1, std::memory_order_acquire);
 
 		if (rw_lock_s_lock_low(lock, pass, file_name, line)) {
 
@@ -416,10 +412,10 @@ rw_lock_x_lock_wait_func(
 	sync_array_t*	sync_arr;
 	int64_t		count_os_wait = 0;
 
-	ut_ad(my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) <= threshold);
+	ut_ad(lock->lock_word <= threshold);
 
 	HMT_low();
-	while (my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) < threshold) {
+	while (lock->lock_word < threshold) {
 		ut_delay(srv_spin_wait_delay);
 
 		if (i < srv_n_spin_wait_rounds) {
@@ -428,7 +424,7 @@ rw_lock_x_lock_wait_func(
 		}
 
 		/* If there is still a reader, then go to sleep.*/
-		++n_spins;
+		n_spins += i;
 
 		sync_cell_t*	cell;
 
@@ -438,8 +434,7 @@ rw_lock_x_lock_wait_func(
 		i = 0;
 
 		/* Check lock_word to ensure wake-up isn't missed.*/
-		if (my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) < threshold) {
-
+		if (lock->lock_word < threshold) {
 			++count_os_wait;
 
 			/* Add debug info as it is needed to detect possible
@@ -528,18 +523,15 @@ rw_lock_x_lock_low(
 					file_name, line);
 
 			} else {
-				int32_t lock_word = my_atomic_load32_explicit(&lock->lock_word,
-									      MY_MEMORY_ORDER_RELAXED);
+				int32_t lock_word = lock->lock_word;
 				/* At least one X lock by this thread already
 				exists. Add another. */
 				if (lock_word == 0
 				    || lock_word == -X_LOCK_HALF_DECR) {
-					my_atomic_add32_explicit(&lock->lock_word, -X_LOCK_DECR,
-								 MY_MEMORY_ORDER_RELAXED);
+					lock->lock_word.fetch_sub(X_LOCK_DECR);
 				} else {
 					ut_ad(lock_word <= -X_LOCK_DECR);
-					my_atomic_add32_explicit(&lock->lock_word, -1,
-								 MY_MEMORY_ORDER_RELAXED);
+					lock->lock_word.fetch_sub(1);
 				}
 			}
 
@@ -611,10 +603,10 @@ rw_lock_sx_lock_low(
 				  read and write to the lock_word. */
 
 #ifdef UNIV_DEBUG
-				int32_t lock_word =
+				auto lock_word =
 #endif
-				my_atomic_add32_explicit(&lock->lock_word, -X_LOCK_HALF_DECR,
-							 MY_MEMORY_ORDER_RELAXED);
+				lock->lock_word.fetch_sub(X_LOCK_HALF_DECR,
+							std::memory_order_relaxed);
 
 				ut_ad((lock_word == 0)
 				      || ((lock_word <= -X_LOCK_DECR)
@@ -662,6 +654,12 @@ rw_lock_x_lock_func(
 	ut_ad(rw_lock_validate(lock));
 	ut_ad(!rw_lock_own(lock, RW_LOCK_S));
 
+	if (rw_lock_x_lock_low(lock, pass, file_name, line)) {
+		/* Locking succeeded */
+		return;
+	}
+	rw_lock_stats.rw_x_spin_wait_count.inc();
+
 lock_loop:
 
 	if (rw_lock_x_lock_low(lock, pass, file_name, line)) {
@@ -681,14 +679,15 @@ lock_loop:
 
 		/* Spin waiting for the lock_word to become free */
 		HMT_low();
+		ulint j = i;
 		while (i < srv_n_spin_wait_rounds
-		       && my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) <= X_LOCK_HALF_DECR) {
+		       && lock->lock_word <= X_LOCK_HALF_DECR) {
 			ut_delay(srv_spin_wait_delay);
 			i++;
 		}
 
 		HMT_medium();
-		spin_count += lint(i);
+		spin_count += lint(i - j);
 
 		if (i >= srv_n_spin_wait_rounds) {
 
@@ -707,7 +706,7 @@ lock_loop:
 
 	/* Waiters must be set before checking lock_word, to ensure signal
 	is sent. This could lead to a few unnecessary wake-up signals. */
-	my_atomic_fas32_explicit(&lock->waiters, 1, MY_MEMORY_ORDER_ACQUIRE);
+	lock->waiters.exchange(1, std::memory_order_acquire);
 
 	if (rw_lock_x_lock_low(lock, pass, file_name, line)) {
 		sync_array_free_cell(sync_arr, cell);
@@ -756,11 +755,17 @@ rw_lock_sx_lock_func(
 	sync_array_t*	sync_arr;
 	lint		spin_count = 0;
 	int64_t		count_os_wait = 0;
-	lint		spin_wait_count = 0;
 
 	ut_ad(rw_lock_validate(lock));
 	ut_ad(!rw_lock_own(lock, RW_LOCK_S));
 
+	if (rw_lock_sx_lock_low(lock, pass, file_name, line)) {
+		/* Locking succeeded */
+		return;
+	}
+
+	rw_lock_stats.rw_sx_spin_wait_count.inc();
+
 lock_loop:
 
 	if (rw_lock_sx_lock_low(lock, pass, file_name, line)) {
@@ -772,23 +777,21 @@ lock_loop:
 		}
 
 		rw_lock_stats.rw_sx_spin_round_count.add(spin_count);
-		rw_lock_stats.rw_sx_spin_wait_count.add(spin_wait_count);
 
 		/* Locking succeeded */
 		return;
 
 	} else {
 
-		++spin_wait_count;
-
 		/* Spin waiting for the lock_word to become free */
+		ulint j = i;
 		while (i < srv_n_spin_wait_rounds
-		       && my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) <= X_LOCK_HALF_DECR) {
+		       && lock->lock_word <= X_LOCK_HALF_DECR) {
 			ut_delay(srv_spin_wait_delay);
 			i++;
 		}
 
-		spin_count += lint(i);
+		spin_count += lint(i - j);
 
 		if (i >= srv_n_spin_wait_rounds) {
 
@@ -807,7 +810,7 @@ lock_loop:
 
 	/* Waiters must be set before checking lock_word, to ensure signal
 	is sent. This could lead to a few unnecessary wake-up signals. */
-	my_atomic_fas32_explicit(&lock->waiters, 1, MY_MEMORY_ORDER_ACQUIRE);
+	lock->waiters.exchange(1, std::memory_order_acquire);
 
 	if (rw_lock_sx_lock_low(lock, pass, file_name, line)) {
 
@@ -820,7 +823,6 @@ lock_loop:
 		}
 
 		rw_lock_stats.rw_sx_spin_round_count.add(spin_count);
-		rw_lock_stats.rw_sx_spin_wait_count.add(spin_wait_count);
 
 		/* Locking succeeded */
 		return;
@@ -846,15 +848,11 @@ rw_lock_validate(
 /*=============*/
 	const rw_lock_t*	lock)	/*!< in: rw-lock */
 {
-	int32_t	lock_word;
-
 	ut_ad(lock);
 
-	lock_word = my_atomic_load32_explicit(const_cast<int32_t*>(&lock->lock_word),
-					      MY_MEMORY_ORDER_RELAXED);
+	int32_t lock_word = lock->lock_word;
 
-	ut_ad(my_atomic_load32_explicit(const_cast<int32_t*>(&lock->waiters),
-					MY_MEMORY_ORDER_RELAXED) < 2);
+	ut_ad(lock->waiters < 2);
 	ut_ad(lock_word > -(2 * X_LOCK_DECR));
 	ut_ad(lock_word <= X_LOCK_DECR);
 
@@ -917,8 +915,7 @@ rw_lock_add_debug_info(
 	rw_lock_debug_mutex_exit();
 
 	if (pass == 0 && lock_type != RW_LOCK_X_WAIT) {
-		int32_t lock_word = my_atomic_load32_explicit(&lock->lock_word,
-							      MY_MEMORY_ORDER_RELAXED);
+		int32_t lock_word = lock->lock_word;
 
 		/* Recursive x while holding SX
 		(lock_type == RW_LOCK_X && lock_word == -X_LOCK_HALF_DECR)
@@ -994,13 +991,22 @@ rw_lock_own(
 	ut_ad(lock);
 	ut_ad(rw_lock_validate(lock));
 
+	const os_thread_id_t thread_id = os_thread_get_curr_id();
+
+	if (!os_thread_eq(lock->writer_thread, thread_id)) {
+	} else if (lock_type == RW_LOCK_X && rw_lock_get_x_lock_count(lock)) {
+		return TRUE;
+	} else if (lock_type == RW_LOCK_SX && rw_lock_get_sx_lock_count(lock)) {
+		return TRUE;
+	}
+
 	rw_lock_debug_mutex_enter();
 
 	for (const rw_lock_debug_t* info = UT_LIST_GET_FIRST(lock->debug_list);
 	     info != NULL;
 	     info = UT_LIST_GET_NEXT(list, info)) {
 
-		if (os_thread_eq(info->thread_id, os_thread_get_curr_id())
+		if (os_thread_eq(info->thread_id, thread_id)
 		    && info->pass == 0
 		    && info->lock_type == lock_type) {
 
@@ -1025,12 +1031,23 @@ bool rw_lock_own_flagged(const rw_lock_t* lock, rw_lock_flags_t flags)
 {
 	ut_ad(rw_lock_validate(lock));
 
+	const os_thread_id_t thread_id = os_thread_get_curr_id();
+
+	if (!os_thread_eq(lock->writer_thread, thread_id)) {
+	} else if ((flags & RW_LOCK_FLAG_X)
+		   && rw_lock_get_x_lock_count(lock)) {
+		return true;
+	} else if ((flags & RW_LOCK_FLAG_SX)
+		   && rw_lock_get_sx_lock_count(lock)) {
+		return true;
+	}
+
 	rw_lock_debug_mutex_enter();
 
 	for (rw_lock_debug_t* info = UT_LIST_GET_FIRST(lock->debug_list);
 	     info != NULL;
 	     info = UT_LIST_GET_NEXT(list, info)) {
-		if (!os_thread_eq(info->thread_id, os_thread_get_curr_id())
+		if (!os_thread_eq(info->thread_id, thread_id)
 		    || info->pass) {
 			continue;
 		}
@@ -1084,11 +1101,11 @@ rw_lock_list_print_info(
 
 		count++;
 
-		if (my_atomic_load32_explicit(const_cast<int32_t*>(&lock->lock_word), MY_MEMORY_ORDER_RELAXED) != X_LOCK_DECR) {
+		if (lock->lock_word != X_LOCK_DECR) {
 
 			fprintf(file, "RW-LOCK: %p ", (void*) lock);
 
-			if (int32_t waiters= my_atomic_load32_explicit(const_cast<int32_t*>(&lock->waiters), MY_MEMORY_ORDER_RELAXED)) {
+			if (int32_t waiters= lock->waiters) {
 				fprintf(file, " (%d waiters)\n", waiters);
 			} else {
 				putc('\n', file);
@@ -1152,10 +1169,10 @@ rw_lock_debug_print(
 	fprintf(f, "\n");
 }
 
-/** Print where it was locked from
+/** Print the rw-lock information.
 @return the string representation */
 std::string
-rw_lock_t::locked_from() const
+rw_lock_t::to_string() const
 {
 	/* Note: For X locks it can be locked form multiple places because
 	the same thread can call X lock recursively. */
@@ -1165,6 +1182,11 @@ rw_lock_t::locked_from() const
 
 	ut_ad(rw_lock_validate(this));
 
+	msg << "RW-LATCH: "
+	    << "thread id " << os_thread_pf(os_thread_get_curr_id())
+	    << " addr: " << this
+	    << " Locked from: ";
+
 	rw_lock_debug_mutex_enter();
 
 	for (rw_lock_debug_t* info = UT_LIST_GET_FIRST(debug_list);
@@ -1187,19 +1209,4 @@ rw_lock_t::locked_from() const
 
 	return(msg.str());
 }
-
-/** Print the rw-lock information.
-@return the string representation */
-std::string
-rw_lock_t::to_string() const
-{
-	std::ostringstream	msg;
-
-	msg << "RW-LATCH: "
-	    << "thread id " << os_thread_pf(os_thread_get_curr_id())
-	    << " addr: " << this
-	    << " Locked from: " << locked_from().c_str();
-
-	return(msg.str());
-}
 #endif /* UNIV_DEBUG */
