@@ -160,6 +160,9 @@ trx_rollback_to_savepoint_low(
 
 	mem_heap_free(heap);
 
+	/* There might be work for utility threads.*/
+	srv_active_wake_master_thread();
+
 	MONITOR_DEC(MONITOR_TRX_ACTIVE);
 }
 
@@ -831,7 +834,7 @@ discard:
         trx_free_at_shutdown(trx);
       }
       else
-        trx_free(trx);
+        trx->free();
     }
   }
 }
